#ifndef POKEMONSCRIPTTESTER_TESTFUNCTIONS_HPP
#define POKEMONSCRIPTTESTER_TESTFUNCTIONS_HPP

#include <angelscript.h>
#include "../RequirementFailed.hpp"
#include "../TestEnvironment.hpp"

class TestFunctions {
private:
    struct RequirementData {
        std::string FileName;
        size_t Line;
    };

    static RequirementData GetRequirementData(asIScriptContext* ctx) {
        auto filename = ctx->GetFunction()->GetScriptSectionName();
        auto line = ctx->GetLineNumber();
        return RequirementData{
            .FileName = filename,
            .Line = (size_t)line,
        };
    }

    static bool Require(bool value) {
        auto* ctx = asGetActiveContext();
        TestEnvironment* env = static_cast<TestEnvironment*>(ctx->GetUserData(684));
        env->TotalRequirements += 1;
        if (!value) {
            env->FailedRequirements += 1;
            auto data = GetRequirementData(ctx);
            throw RequirementFailed(data.FileName, data.Line);
        }
        return true;
    }

#define GetEqualityMessage(expected, actual)                                                                           \
    std::string eqMsg;                                                                                                 \
    {                                                                                                                  \
        std::ostringstream ss;                                                                                         \
        ss << "Expected: '" << std::to_string(expected) << "', but was: '" << std::to_string(actual) << "'";           \
        eqMsg = ss.str();                                                                                              \
    }

    static bool RequireEqualsI32(i32 expected, i32 actual) {
        auto* ctx = asGetActiveContext();
        TestEnvironment* env = static_cast<TestEnvironment*>(ctx->GetUserData(684));
        env->TotalRequirements += 1;
        if (expected != actual) {
            env->FailedRequirements += 1;
            auto data = GetRequirementData(ctx);
            GetEqualityMessage(expected, actual);
            throw RequirementFailed(data.FileName, data.Line, eqMsg);
        }
        return true;
    }

    static bool RequireEqualsString(const std::string& expected, const std::string& actual) {
        auto* ctx = asGetActiveContext();
        TestEnvironment* env = static_cast<TestEnvironment*>(ctx->GetUserData(684));
        if (expected != actual) {
            env->FailedRequirements += 1;
            auto data = GetRequirementData(ctx);
            std::string eqMsg;
            {
                std::ostringstream ss;
                ss << "Expected: '" << expected << "', but was: '" << actual << "'";
                eqMsg = ss.str();
            }
            throw RequirementFailed(data.FileName, data.Line, eqMsg);
        }
        return true;
    }

public:
    static void Register(AngelScriptResolver* scriptResolver) {
        auto engine = scriptResolver->GetBuilder().GetEngine();
        Ensure(engine->RegisterGlobalFunction("bool Require(bool expression)", asFUNCTION(Require), asCALL_CDECL) >= 0);
        Ensure(engine->RegisterGlobalFunction("bool RequireEquals(int expected, int actual)",
                                              asFUNCTION(RequireEqualsI32), asCALL_CDECL) >= 0);
        Ensure(engine->RegisterGlobalFunction("bool RequireEquals(const string &in expected, const string &in actual)",
                                              asFUNCTION(RequireEqualsString), asCALL_CDECL) >= 0);
    }
};

#endif // POKEMONSCRIPTTESTER_TESTFUNCTIONS_HPP
